import { toReactive } from "@vueuse/core";
import { v4 } from "uuid";
import { computed, isReactive, reactive } from "vue";
export default (tree, { branch: keyBranch = "branch", children: keyChildren = "children", id: keyId = "id", index: keyIndex = "index", next: keyNext = "next", parent: keyParent = "parent", prev: keyPrev = "prev", siblings: keySiblings = "siblings", } = {}) => {
    const properties = {
        [keyBranch]: {
            get() {
                const ret = [this];
                while (ret[0]?.[keyParent])
                    ret.unshift(ret[0][keyParent]);
                return ret;
            },
        },
        [keyIndex]: {
            get() {
                return this[keySiblings].findIndex((sibling) => this[keyId] === sibling[keyId]);
            },
        },
        [keyNext]: {
            get() {
                return this[keySiblings][this[keyIndex] + 1];
            },
        },
        [keyPrev]: {
            get() {
                return this[keySiblings][this[keyIndex] - 1];
            },
        },
    };
    const getLeaves = (siblings, parent = {}) => siblings.value.flatMap((value) => {
        Object.defineProperties(value, {
            ...properties,
            [keyParent]: parent,
            [keySiblings]: siblings,
        });
        return [
            value,
            ...getLeaves({
                configurable: true,
                value: (value[keyChildren] ?? []),
            }, { configurable: true, value }),
        ];
    }), leaves = computed(() => getLeaves({ value: isReactive(tree) ? tree : reactive(tree) })), objLeaves = toReactive(computed(() => Object.fromEntries(leaves.value.map((leaf) => [leaf[keyId], leaf]))));
    return {
        add: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const children = the[keyChildren], index = the[keyIndex], siblings = the[keySiblings];
                const id = v4();
                switch (true) {
                    case !!the[keyParent]:
                        siblings.splice(index + 1, 0, { [keyId]: id });
                        break;
                    case !!children:
                        children.unshift({ [keyId]: id });
                        break;
                    default:
                        siblings.splice(index + 1, 0, { [keyId]: id });
                        break;
                }
                return id;
            }
            return undefined;
        },
        arrLeaves: toReactive(leaves),
        down: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const index = the[keyIndex], nextIndex = index + 1, siblings = the[keySiblings];
                if (index < siblings.length - 1 &&
                    siblings[index] &&
                    siblings[nextIndex])
                    [siblings[index], siblings[nextIndex]] = [
                        siblings[nextIndex],
                        siblings[index],
                    ];
            }
        },
        leaves,
        left: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const parent = the[keyParent];
                if (parent?.[keyParent]) {
                    const children = (parent[keyChildren] ?? []), siblings = parent[keySiblings];
                    siblings.splice(parent[keyIndex] + 1, 0, ...children.splice(the[keyIndex], 1));
                    return parent[keyId];
                }
            }
            return undefined;
        },
        objLeaves,
        remove: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const parent = the[keyParent];
                if (parent) {
                    const [root] = leaves.value, next = the[keyNext], prev = the[keyPrev], id = (next?.[keyId] ??
                        prev?.[keyId] ??
                        parent[keyId] ??
                        root?.[keyId]), siblings = the[keySiblings];
                    siblings.splice(the[keyIndex], 1);
                    return id;
                }
            }
            return undefined;
        },
        right: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const prev = the[keyPrev];
                if (prev) {
                    const children = (prev[keyChildren] ?? []), id = prev[keyId], siblings = the[keySiblings];
                    prev[keyChildren] = [
                        ...children,
                        ...siblings.splice(the[keyIndex], 1),
                    ];
                    return id;
                }
            }
            return undefined;
        },
        up: (pId) => {
            const the = objLeaves[pId];
            if (the) {
                const index = the[keyIndex], prevIndex = index - 1, siblings = the[keySiblings];
                if (index && siblings[index] && siblings[prevIndex])
                    [siblings[prevIndex], siblings[index]] = [
                        siblings[index],
                        siblings[prevIndex],
                    ];
            }
        },
    };
};
